# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from unittest import TestCase


class TemplateLibraryTestCase(TestCase):

    SETTINGS_ENVNAME = "DJANGO_SETTINGS_MODULE"

    def setUp(self):
        self.settings = os.environ.get(self.SETTINGS_ENVNAME, "")
        os.environ[self.SETTINGS_ENVNAME] = "tests.kit_settings"

    def tearDown(self):
        os.environ[self.SETTINGS_ENVNAME] = self.settings

    def test_directtag(self):
        from django.template import Context
        from django.template.engine import Engine

        from neom.kit.template.library import Library

        library = Library()

        @library.directtag
        def dummy(foo: str, bar: str = "bar"):
            return f"{foo}-{bar}"

        engine = Engine()
        engine.template_libraries["test_library"] = library

        html = engine.from_string(
            "{% load test_library %}{% dummy 'foo' %}{% dummy 'oof' 'baz' %}"
        ).render(Context())

        self.assertEqual(html, "foo-baroof-baz")



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Staff tests."""

from __future__ import annotations

from unittest import TestCase

from neom.new_ddd import staff


class PhoneTestCase(TestCase):
    def test_creation(self):
        phone = staff.Phone(country=51, area=123, number=1234567)

        self.assertEqual(51, phone.country)
        self.assertEqual(123, phone.area)
        self.assertEqual(1234567, phone.number)


class MobileTestCase(TestCase):
    def test_creation(self):
        mobile = staff.Mobile.Make("(+51) 987-654-321")
        self.assertEqual(mobile.country, 51)
        self.assertEqual(mobile.area, None)
        self.assertEqual(mobile.number, 987654321)

    def test_format(self):
        mobile = staff.Mobile.Make("(+51) 987-654-321")
        self.assertEqual(str(mobile), "(+51) 987-654-321")

    def test_bad_creation(self):
        with self.assertRaisesRegex(
            ValueError,
            r"Invalid mobile phone \+51987654321. Use \(\+xx\) xxx-xxx-xxx.",
        ):
            staff.Mobile.Make("+51987654321")


class EmailTestCase(TestCase):
    def test_creation_and_format(self):
        email = staff.Email(address="dummy@staff.ui")
        self.assertEqual(str(email), "dummy@staff.ui")

    def test_validate(self):
        email = staff.Email(address="dummy.at.staff.ui")
        with self.assertRaisesRegex(
            ValueError, r"Invalid email address format dummy\.at\.staff\.ui"
        ):
            email.Validate()



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""ValueObject tests."""

from __future__ import annotations

from copy import deepcopy
from datetime import datetime
from typing import ForwardRef, cast
from unittest import TestCase

from neom.new_ddd.shared import Field, ValueObject


class ValueObjectDeclarationTestCase(TestCase):
    """ValueObject declaration test case."""

    def test_primitives(self):
        """Test most common value object definition."""

        class Person(ValueObject[ForwardRef("Person")]):
            """Dummy."""

            name: Field[str]
            age: Field[int]
            birth: Field[datetime]

            def SameValueAs(self, other: Person) -> bool:
                return (
                    self.name == other.name
                    and self.age == other.age
                    and self.birth == other.birth
                )

            def Copy(self) -> ForwardRef("Person"):
                return Person(name=self.name, age=self.age, birth=self.birth)

        person = Person(name="Bruce Wayne", age=9, birth=datetime(2010, 1, 2))

        self.assertIsInstance(person, Person)
        self.assertEqual(person.name, "Bruce Wayne")
        self.assertEqual(person.age, 9)
        self.assertEqual(person.birth, datetime(2010, 1, 2))
        self.assertTrue(person.SameValueAs(person))
        self.assertTrue(person.SameValueAs(person.Copy()))

    def test_inheritance(self):
        """Test value object inheritance."""

        class PersonalInfo(ValueObject):
            """Dummy."""

            name: Field[str]
            age: Field[int]

            def SameValueAs(self, other: PersonalInfo) -> bool:
                return self.name == other.name and self.age == other.age

            def Copy(self) -> ForwardRef("PersonalInfo"):
                return PersonalInfo(name=self.name, age=self.age)

        class Person(PersonalInfo):
            """Dummy."""

            birth: Field[datetime]

            def SameValueAs(self, other: Person) -> bool:
                return (
                    super().SameValueAs(cast(PersonalInfo, other))
                    and self.birth == other.birth
                )

            def Copy(self) -> ForwardRef("Person"):
                personalInfo = super().Copy()
                return Person(
                    name=personalInfo.name,
                    age=personalInfo.age,
                    birth=self.birth,
                )

        person = Person(name="Bruce Wayne", age=9, birth=datetime(2010, 1, 2))

        self.assertIsInstance(person, Person)
        self.assertEqual(person.name, "Bruce Wayne")
        self.assertEqual(person.age, 9)
        self.assertEqual(person.birth, datetime(2010, 1, 2))
        self.assertTrue(person.SameValueAs(person))
        self.assertTrue(person.SameValueAs(person.Copy()))


class ValueObjectMethodsTestCase(TestCase):
    """ValueObject methods test case."""

    def setUp(self):
        class Person(ValueObject[ForwardRef("Person")]):
            """Dummy."""

            name: Field[str]
            age: Field[int]
            birth: Field[datetime]

            def SameValueAs(self, other: Person) -> bool:
                return (
                    self.name == other.name
                    and self.age == other.age
                    and self.birth == other.birth
                )

            def Copy(self) -> ForwardRef("Person"):
                return deepcopy(self)

        self.person = Person

    def test_same_value_as(self):
        """Test SameValueAs method."""
        p1 = self.person(name="Bruce Wayne", age=9, birth=datetime(2010, 1, 2))
        p2 = self.person(name="Bruce Wayne", age=9, birth=datetime(2010, 1, 2))
        self.assertTrue(p1.SameValueAs(p2))

    def test_copy(self):
        """Test Copy method."""
        p1 = self.person(name="Bruce Wayne", age=9, birth=datetime(2010, 1, 2))
        p2 = p1.Copy()

        # TODO: improve this test to verify deep copy

        self.assertIs(p1.name, p2.name)
        self.assertEqual(p1.name, p2.name)

        self.assertIs(p1.age, p2.age)
        self.assertEqual(p1.age, p2.age)

        self.assertIsNot(p1.birth, p2.birth)
        self.assertEqual(p1.birth, p2.birth)


class ValueObjectBackCompatibilityTestCase(TestCase):
    """ValueObject back compatibility test case."""

    def test_auto_fields(self):
        """Test SameValueAs method."""

        class Person(ValueObject[ForwardRef("Person")]):
            """Dummy."""

            name: str
            age: int
            birth: datetime

            def SameValueAs(self, other: Person) -> bool:
                return (
                    self.name == other.name
                    and self.age == other.age
                    and self.birth == other.birth
                )

            def Copy(self) -> ForwardRef("Person"):
                return Person(name=self.name, age=self.age, birth=self.birth)

        person = Person(name="Bruce Wayne", age=9, birth=datetime(2010, 1, 2))

        self.assertIsInstance(person, Person)
        self.assertEqual(person.name, "Bruce Wayne")
        self.assertEqual(person.age, 9)
        self.assertEqual(person.birth, datetime(2010, 1, 2))
        self.assertTrue(person.SameValueAs(person))
        self.assertTrue(person.SameValueAs(person.Copy()))



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Entity tests."""

from __future__ import annotations

from datetime import datetime
from typing import ForwardRef
from unittest import TestCase

from neom.new_ddd.shared import Entity, Field


class EntityDeclarationTestCase(TestCase):
    """Entity declaration test case."""

    def test_definition(self):
        """Test most common entity definition."""

        class Person(Entity[ForwardRef("Person"), str]):
            """Dummy."""

            name: Field[str]
            age: Field[int]
            birth: Field[datetime]

            def Identity(self) -> str:
                return self.name

            def SameIdentityAs(self, other: Person) -> bool:
                return self.Identity() == other.Identity()

        person = Person(name="Bruce Banner", age=3, birth=datetime(2000, 10, 1))

        self.assertIsInstance(person, Person)
        self.assertEqual(person.name, "Bruce Banner")
        self.assertEqual(person.age, 3)
        self.assertEqual(person.birth, datetime(2000, 10, 1))
        self.assertEqual(person.Identity(), "Bruce Banner")
        self.assertTrue(person.SameIdentityAs(person))



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import django.views.generic.edit as edit_views
from django.core.exceptions import ImproperlyConfigured
from django.db.models import fields as model_fields
from django.forms import fields as form_fields
from django.forms import models as model_forms

from neom.kit.md2.forms import models as md2_model_forms
from neom.kit.md2.forms import widgets as md2_widgets


class AttachFieldMixin(form_fields.Field):
    def get_bound_field(self, form, field_name):
        boundfield = super().get_bound_field(form, field_name)
        self.widget.field = boundfield
        return boundfield


class CharField(form_fields.CharField, AttachFieldMixin):
    pass


class TypedChoiceField(form_fields.TypedChoiceField, AttachFieldMixin):
    pass


def formfield_callback(field, **kwargs):
    if isinstance(field, model_fields.CharField):
        return CharField(**kwargs, widget=md2_widgets.TextInput)

    if isinstance(field, model_fields.IntegerField):
        if field.choices:
            return TypedChoiceField(
                choices=field.choices, **kwargs, widget=md2_widgets.Select
            )
        # else:
        # return form_fields.IntegerField(...)

    return field.formfield(**kwargs)


class UpdateView(edit_views.UpdateView):
    def get_form_class(self):
        if self.fields is not None and self.form_class:
            raise ImproperlyConfigured(
                "Only specify one of 'fields' or 'form_class'."
            )
        if self.form_class:
            return self.form_class
        else:
            model = (
                self.model
                if self.model is not None
                else self.get_queryset().model
            )

            if self.fields is None:
                raise ImproperlyConfigured(
                    "Using without the 'fields' attribute is prohibited."
                )

            return model_forms.modelform_factory(
                model,
                md2_model_forms.ModelForm,
                fields=self.fields,
                formfield_callback=formfield_callback,
            )



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Stuff definitions for enitty and stuffs designing."""

from __future__ import annotations

import builtins
from abc import ABC, ABCMeta
from sys import modules
from types import FunctionType
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    NamedTuple,
    Tuple,
    Type,
    TypeVar,
    get_type_hints,
)

from .exceptions import BadMemberError

__all__ = ("Field",)

T = TypeVar("T")


class KindnameError(BadMemberError):
    """Raised when a member definition for a kindname can not be found. May also
    be raised when the kind is not a string."""


class _MetaStuff(ABCMeta):
    """Metaclass for Stuff

    To state members with the names needed. For example,

    .. code-block: python
      class Publication(Stuff):
        count: int

    ``Publication.count`` needs to know the name ``count`` defined.
    """

    def __init__(cls, name: str, bases: Tuple[str], namespace: Dict[str, Any]):
        super(_MetaStuff, cls).__init__(name, bases, namespace)
        cls._StateMembers()

    def __repr__(cls):
        fields = [f"{f.name}={f!r}" for _, f in sorted(cls._fields.items())]
        return f'{cls.__name__}<{", ".join(fields)}>'


class Stuff(ABC, metaclass=_MetaStuff):
    """A class describing domain stuffs.

    All stuff classes inheriting from :class:`Stuff` automatically have
    :class:`_MetaStuff` as their metaclass, so that members are stated
    properly after class definition."""

    __slots__ = ()

    _fields = {}
    _kinds = {}

    def __init__(self, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def __repr__(self):
        fields = [f"{f.name}={f!r}" for _, f in sorted(self._fields.items())]
        return f'{self._kindname()}<{", ".join(fields)}>'

    def ReflectionEquals(self, other: Stuff) -> bool:
        """Compare two stuff by fields."""
        if len(self._fields) != len(other._fields):
            return False
        return not any(
            name not in other._fields
            or field != other._fields[name]
            or getattr(self, name) != getattr(other, name)
            for name, field in self._fields.items()
        )

    @classmethod
    def _kindname(cls):
        """Return the kind name for this class.

        Defaults to ``cls.__name__``; users may override this to return a
        different name when stored than the name of the class.
        """
        return cls.__name__

    @classmethod
    def _StateMembers(cls):
        """State member calling their ``_FixState()`` method.

        .. note::

          Called by :class:`_MetaStuff`, but may also be called
          manually after dynamically updating a stuff class.

        Raises:
          KindnameError: When returned kind name from ``_kindname()`` is not a
            :class:`str`.
        """
        kindname = cls._kindname()
        if not isinstance(kindname, str):
            raise KindnameError(
                f"Class {cls.__name__} defines a ``_kindname()`` method that"
                f" returns a non-string ({kindname!r})"
            )

        typeHints = get_type_hints(cls)
        if typeHints:
            cls._fields = {}
            cls._kinds = {}

            for name, field in typeHints.items():
                # TODO: back compatibility - remove after migration
                if not isinstance(field, Field):
                    field = Field[field]
                field._FixState(cls, name)  # pylint:disable=protected-access
                cls._fields[name] = field
            cls._UpdateKinds()

            if (
                hasattr(cls, "Meta")
                and hasattr(cls.Meta, "__init__")
                and isinstance(cls.Meta.__init__, bool)
                and cls.Meta.__init__
            ):
                _StateNewAttribute(
                    cls,
                    "__init__",
                    _InitFn(
                        cls._fields.values(),
                        "self",
                        modules[cls.__module__].__dict__
                        if cls.__module__ in modules
                        else {},
                    ),
                )

    @classmethod
    def _UpdateKinds(cls: Type[Stuff]):
        """Update kinds dict to include this class."""
        cls._kinds[cls._kindname()] = cls


def _InitFn(
    fields: List[Field], selfname: str, globalns: Dict[str, Type[object]]
):
    """Init Function."""
    localns = {f"_type_{f.name}": f.kind for f in fields}
    lines = [f"{selfname}.{f.name}={f.name}" for f in fields]
    params = [f"{f.name}:{f.kind.__name__}" for f in fields]

    return _NewFn(
        "__init__",
        [selfname] + params,
        lines,
        localns=localns,
        globalns=globalns,
        return_type=bool,
    )


def _StateNewAttribute(
    cls: Type[object], name: str, value: Callable[..., None]
):
    """Never overwrites an existing member."""
    if name in cls.__dict__:
        raise ValueError(f"{cls}: has attribute name={name} with value={value}")
    _StateQualname(cls, value)
    setattr(cls, name, value)


def _StateQualname(cls, value):
    """_NewFn uses proper __qualname__."""
    if isinstance(value, FunctionType):
        value.__qualname__ = f"{cls.__qualname__}.{value.__name__}"
    return value


def _NewFn(
    name: str,
    args: List[str],
    body: List[str],
    *,
    globalns: Dict[str, Type[object]] = None,
    localns: Dict[str, Type[object]] = None,
    return_type: Type[object] = None,
):
    """Create new function. localns mutates on exec()."""
    if localns is None:
        localns = {}
    if "BUILTINS" not in localns:
        localns["BUILTINS"] = builtins
    return_annotation = ""
    if return_type:
        localns["_return_type"] = return_type
        return_annotation = "->_return_type"
    args = ",".join(args)
    body = "\n".join(f"  {b}" for b in body)
    txt = f" def {name}({args}){return_annotation}:\n{body}"
    localvars = ", ".join(localns.keys())
    txt = f"def __new_fn__({localvars}):\n{txt}\n return {name}"
    ns = {}
    exec(txt, globalns, ns)  # pylint:disable=exec-used
    return ns["__new_fn__"](**localns)


class Field(Generic[T]):
    """A class describing a persisted kind member."""

    __slots__ = ("kind", "validator", "pkind", "name")

    def __init__(self, kind: Type[T], validator: Callable[[T], None] = None):
        self.kind = kind
        self.validator = validator
        self.name = None
        self.pkind = None

    def __repr__(self):
        return (
            f"Field[{self.kind} name={self.name},"
            f" pkind={self.pkind._kindname()}]"
        )

    def __call__(self, *args, **kwds):
        raise TypeError(f"Cannot instantiate {self!r}")

    def __eq__(self, other: Field) -> bool:
        return (
            other
            and isinstance(other, Field)
            and self.kind is other.kind
            and self.name == other.name
        )

    def _FixState(self, pkind: Type[T], name: str):
        """Helper called to tell member its name."""
        self.pkind = pkind
        self.name = name

    def __class_getitem__(
        cls, item: T | Tuple[Type[T], Callable[[T], None] | Arg]
    ):
        return cls(*item) if isinstance(item, Arg) else cls(item)


class Arg(NamedTuple):
    """To describe Field definition."""

    kind: Type[T]
    validator: Callable[[T], None] = None



# Copyright 2022 neomadas-dev
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#   2. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#   3. Neither the name of the copyright holder nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
from pathlib import Path

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    help = (
        "Delete current db and recreate migrations to execute population data."
    )

    def handle(self, *unused_args, **unsed_options):
        basedir = Path(settings.BASE_DIR)

        if not basedir.is_dir():
            raise CommandError(f"Invalid project directory: {basedir}")

        dbpath = Path(settings.DATABASES["default"]["NAME"])
        if dbpath.exists():
            dbpath.unlink()
        else:
            logging.warning(f"db path {dbpath} doesn't exist")

        call_command("makemigrations")
        call_command("migrate")
        call_command("db-populate")
